@@ -1,3 +1,5 @@
+from __future__ import absolute_import, division, print_function
+
 import os
 
 import numpy as np
@@ -5,7 +7,7 @@ import numpy as np
 from qtpy import QtWidgets, QtGui
 from qtpy.QtCore import Qt
 
-from glue.external.echo.qt import autoconnect_callbacks_to_qt
+from glue.external.echo.qt import autoconnect_callbacks_to_qt, connect_value
 from glue.utils.qt import load_ui, fix_tab_widget_fontsize
 
 
@@ -20,8 +22,12 @@ class ScatterLayerStyleEditor(QtWidgets.QWidget):
 
         connect_kwargs = {'alpha': dict(value_range=(0, 1)),
                           'size_scaling': dict(value_range=(0.1, 10), log=True),
+                          'density_contrast': dict(value_range=(0, 1)),
                           'vector_scaling': dict(value_range=(0.1, 10), log=True)}
-        autoconnect_callbacks_to_qt(layer.state, self.ui, connect_kwargs)
+        self._connections = autoconnect_callbacks_to_qt(layer.state, self.ui, connect_kwargs)
+
+        self._connection_dpi = connect_value(layer.state.viewer_state, 'dpi', self.ui.value_dpi,
+                                             value_range=(12, 144), log=True)
 
         fix_tab_widget_fontsize(self.ui.tab_widget)
 
@@ -37,6 +43,9 @@ class ScatterLayerStyleEditor(QtWidgets.QWidget):
         self.layer_state.add_callback('size_mode', self._update_size_mode)
         self.layer_state.add_callback('vector_mode', self._update_vector_mode)
 
+        self.layer_state.add_callback('density_map', self._update_size_mode)
+        self.layer_state.add_callback('density_map', self._update_warnings)
+
         self.layer_state.add_callback('layer', self._update_warnings)
 
         self._update_markers_visible()
@@ -51,7 +60,7 @@ class ScatterLayerStyleEditor(QtWidgets.QWidget):
 
         self._update_warnings()
 
-    def _update_warnings(self):
+    def _update_warnings(self, *args):
 
         if self.layer_state.layer is None:
             n_points = 0
@@ -63,7 +72,7 @@ class ScatterLayerStyleEditor(QtWidgets.QWidget):
         for combo, threshold in [(self.ui.combosel_size_mode, 10000),
                                  (self.ui.combosel_cmap_mode, 50000)]:
 
-            if n_points > threshold:
+            if n_points > threshold and not self.layer_state.density_map:
                 for item in range(combo.count()):
                     text = combo.itemText(item)
                     if text != 'Fixed':
@@ -89,22 +98,30 @@ class ScatterLayerStyleEditor(QtWidgets.QWidget):
 
     def _update_size_mode(self, size_mode=None):
 
-        if self.layer_state.size_mode == 'Fixed':
-            self.ui.label_size_attribute.hide()
-            self.ui.combosel_size_att.hide()
-            self.ui.label_size_limits.hide()
-            self.ui.valuetext_size_vmin.hide()
-            self.ui.valuetext_size_vmax.hide()
-            self.ui.button_flip_size.hide()
-            self.ui.value_size.show()
-        else:
-            self.ui.label_size_attribute.show()
-            self.ui.combosel_size_att.show()
-            self.ui.label_size_limits.show()
-            self.ui.valuetext_size_vmin.show()
-            self.ui.valuetext_size_vmax.show()
-            self.ui.button_flip_size.show()
-            self.ui.value_size.hide()
+        visible = not self.layer_state.density_map and not self.layer_state.size_mode == 'Fixed'
+        self.ui.label_size_attribute.setVisible(visible)
+        self.ui.combosel_size_att.setVisible(visible)
+        self.ui.label_size_limits.setVisible(visible)
+        self.ui.valuetext_size_vmin.setVisible(visible)
+        self.ui.valuetext_size_vmax.setVisible(visible)
+        self.ui.button_flip_size.setVisible(visible)
+
+        visible = not self.layer_state.density_map and self.layer_state.size_mode == 'Fixed'
+        self.ui.value_size.setVisible(visible)
+
+        density = self.layer_state.density_map
+        self.ui.value_dpi.setVisible(density)
+        self.ui.label_dpi.setVisible(density)
+        self.ui.label_stretch.setVisible(density)
+        self.ui.combosel_stretch.setVisible(density)
+        self.ui.value_density_contrast.setVisible(density)
+        self.ui.label_contrast.setVisible(density)
+        self.ui.combosel_size_mode.setVisible(not density)
+        self.ui.value_size_scaling.setVisible(not density)
+        self.ui.label_size_mode.setVisible(not density)
+        self.ui.label_size_scaling.setVisible(not density)
+        self.ui.label_fill.setVisible(not density)
+        self.ui.bool_fill.setVisible(not density)
 
     def _update_markers_visible(self, *args):
         self.ui.combosel_size_mode.setEnabled(self.layer_state.markers_visible)
@@ -114,6 +131,11 @@ class ScatterLayerStyleEditor(QtWidgets.QWidget):
         self.ui.valuetext_size_vmax.setEnabled(self.layer_state.markers_visible)
         self.ui.button_flip_size.setEnabled(self.layer_state.markers_visible)
         self.ui.value_size_scaling.setEnabled(self.layer_state.markers_visible)
+        self.ui.value_dpi.setEnabled(self.layer_state.markers_visible)
+        self.ui.combosel_stretch.setEnabled(self.layer_state.markers_visible)
+        self.ui.label_size_scaling.setEnabled(self.layer_state.markers_visible)
+        self.ui.combosel_points_mode.setEnabled(self.layer_state.markers_visible)
+        self.ui.value_density_contrast.setEnabled(self.layer_state.markers_visible)
 
     def _update_line_visible(self, *args):
         self.ui.value_linewidth.setEnabled(self.layer_state.line_visible)
