@@ -9,19 +9,17 @@ from matplotlib.colors import ColorConverter
 
 from glue.utils.qt import get_qapp
 from glue.config import viewer_tool
-from glue.core.layer_artist import LayerArtistBase
-from glue.core import message as msg
-from glue.core import Data
+from glue.core import BaseData, Data
 from glue.utils.qt import load_ui
 from glue.viewers.common.qt.data_viewer import DataViewer
 from glue.viewers.common.qt.toolbar import BasicToolbar
-from glue.viewers.common.qt.tool import CheckableTool
+from glue.viewers.common.tool import CheckableTool
+from glue.viewers.common.layer_artist import LayerArtist
 from glue.core.subset import ElementSubsetState
-from glue.core.edit_subset_mode import EditSubsetMode
-from glue.core.state import lookup_class_with_patches
 from glue.utils.colors import alpha_blend_colors
-from glue.utils.qt import mpl_to_qt4_color, messagebox_on_error
+from glue.utils.qt import mpl_to_qt_color, messagebox_on_error
 from glue.core.exceptions import IncompatibleAttribute
+from glue.viewers.table.compat import update_table_viewer_state
 
 __all__ = ['TableViewer', 'TableLayerArtist']
 
@@ -36,28 +34,30 @@ class DataTableModel(QtCore.QAbstractTableModel):
             raise ValueError("Can only use Table widget for 1D data")
         self._table_viewer = table_viewer
         self._data = table_viewer.data
-        self.show_hidden = False
+        self.show_coords = False
         self.order = np.arange(self._data.shape[0])
+        self._update_visible()
 
     def data_changed(self):
         top_left = self.index(0, 0)
         bottom_right = self.index(self.columnCount(), self.rowCount())
+        self._update_visible()
         self.dataChanged.emit(top_left, bottom_right)
         self.layoutChanged.emit()
 
     @property
     def columns(self):
-        if self.show_hidden:
+        if self.show_coords:
             return self._data.components
         else:
-            return self._data.visible_components
+            return self._data.main_components + self._data.derived_components
 
     def columnCount(self, index=None):
         return len(self.columns)
 
     def rowCount(self, index=None):
         # Qt bug: Crashes on tables bigger than this
-        return min(self._data.size, 71582788)
+        return min(self.order_visible.size, 71582788)
 
     def headerData(self, section, orientation, role):
         if role != Qt.DisplayRole:
@@ -70,7 +70,7 @@ class DataTableModel(QtCore.QAbstractTableModel):
                 column_name += "\n{0}".format(units)
             return column_name
         elif orientation == Qt.Vertical:
-            return str(self.order[section])
+            return str(self.order_visible[section])
 
     def data(self, index, role):
 
@@ -80,12 +80,8 @@ class DataTableModel(QtCore.QAbstractTableModel):
         if role == Qt.DisplayRole:
 
             c = self.columns[index.column()]
-            idx = self.order[index.row()]
-            comp = self._data.get_component(c)
-            if comp.categorical:
-                comp = comp.labels
-            else:
-                comp = comp.data
+            idx = self.order_visible[index.row()]
+            comp = self._data[c]
             if isinstance(comp[idx], bytes):
                 return comp[idx].decode('ascii')
             else:
@@ -93,12 +89,12 @@ class DataTableModel(QtCore.QAbstractTableModel):
 
         elif role == Qt.BackgroundRole:
 
-            idx = self.order[index.row()]
+            idx = self.order_visible[index.row()]
 
             # Find all subsets that this index is part of
             colors = []
             for layer_artist in self._table_viewer.layers[::-1]:
-                if isinstance(layer_artist.layer, Data):
+                if isinstance(layer_artist.layer, BaseData):
                     continue
                 if layer_artist.visible:
                     subset = layer_artist.layer
@@ -113,35 +109,58 @@ class DataTableModel(QtCore.QAbstractTableModel):
             # Blend the colors using alpha blending
             if len(colors) > 0:
                 color = alpha_blend_colors(colors, additional_alpha=0.5)
-                color = mpl_to_qt4_color(color)
+                color = mpl_to_qt_color(color)
                 return QtGui.QBrush(color)
 
     def sort(self, column, ascending):
         c = self.columns[column]
         comp = self._data.get_component(c)
-        if comp.categorical:
-            self.order = np.argsort(comp.labels)
-        else:
-            self.order = np.argsort(comp.data)
+        self.order = np.argsort(comp.data)
         if ascending == Qt.DescendingOrder:
             self.order = self.order[::-1]
+        self._update_visible()
         self.layoutChanged.emit()
 
+    def _update_visible(self):
+        """
+        Given which layers are visible or not, convert order to order_visible.
+        """
 
-class TableLayerArtist(LayerArtistBase):
+        # First, if the data layer is visible, show all rows
+        for layer_artist in self._table_viewer.layers:
+            if layer_artist.visible and isinstance(layer_artist.layer, BaseData):
+                self.order_visible = self.order
+                return
 
-    def __init__(self, layer, table_viewer):
+        # If not then we need to show only the rows with visible subsets
+        visible = np.zeros(self.order.shape, dtype=bool)
+        for layer_artist in self._table_viewer.layers:
+            if layer_artist.visible:
+                visible |= layer_artist.layer.to_mask()
+
+        self.order_visible = self.order[visible]
+
+
+class TableLayerArtist(LayerArtist):
+
+    def __init__(self, table_viewer, viewer_state, layer_state=None, layer=None):
         self._table_viewer = table_viewer
-        super(TableLayerArtist, self).__init__(layer)
+        super(TableLayerArtist, self).__init__(viewer_state,
+                                               layer_state=layer_state,
+                                               layer=layer)
+        self.redraw()
 
-    def redraw(self):
+    def _refresh(self):
         self._table_viewer.model.data_changed()
 
+    def redraw(self):
+        self._refresh()
+
     def update(self):
-        pass
+        self._refresh()
 
     def clear(self):
-        pass
+        self._refresh()
 
 
 @viewer_tool
@@ -163,6 +182,9 @@ class RowSelectTool(CheckableTool):
         self.viewer.ui.table.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
 
     def deactivate(self):
+        # Don't do anything if the viewer has already been closed
+        if self.viewer is None:
+            return
         self.viewer.ui.table.setSelectionMode(QtWidgets.QAbstractItemView.NoSelection)
         self.viewer.ui.table.clearSelection()
 
@@ -181,11 +203,15 @@ class TableViewer(DataViewer):
     LABEL = "Table Viewer"
 
     _toolbar_cls = BasicToolbar
+    _data_artist_cls = TableLayerArtist
+    _subset_artist_cls = TableLayerArtist
+
+    inherit_tools = False
     tools = ['table:rowselect']
 
-    def __init__(self, session, parent=None, widget=None):
+    def __init__(self, session, state=None, parent=None, widget=None):
 
-        super(TableViewer, self).__init__(session, parent)
+        super(TableViewer, self).__init__(session, state=state, parent=parent)
 
         self.ui = load_ui('data_viewer.ui',
                           directory=os.path.dirname(__file__))
@@ -202,6 +228,9 @@ class TableViewer(DataViewer):
 
         self.ui.table.selection_changed.connect(self.selection_changed)
 
+        self.state.add_callback('layers', self._on_layers_changed)
+        self._on_layers_changed()
+
     def selection_changed(self):
         app = get_qapp()
         if app.queryKeyboardModifiers() == Qt.AltModifier:
@@ -215,9 +244,9 @@ class TableViewer(DataViewer):
 
     def finalize_selection(self, clear=True):
         model = self.ui.table.selectionModel()
-        selected_rows = [self.model.order[x.row()] for x in model.selectedRows()]
+        selected_rows = [self.model.order_visible[x.row()] for x in model.selectedRows()]
         subset_state = ElementSubsetState(indices=selected_rows, data=self.data)
-        mode = EditSubsetMode()
+        mode = self.session.edit_subset_mode
         mode.update(self._data, subset_state, focus_data=self.data)
         if clear:
             # We block the signals here to make sure that we don't update
@@ -226,70 +255,30 @@ class TableViewer(DataViewer):
             self.ui.table.clearSelection()
             self.ui.table.blockSignals(False)
 
-    def register_to_hub(self, hub):
-
-        super(TableViewer, self).register_to_hub(hub)
-
-        def dfilter(x):
-            return x.sender.data is self.data
-
-        hub.subscribe(self, msg.SubsetCreateMessage,
-                      handler=self._refresh,
-                      filter=dfilter)
-
-        hub.subscribe(self, msg.SubsetUpdateMessage,
-                      handler=self._refresh,
-                      filter=dfilter)
-
-        hub.subscribe(self, msg.SubsetDeleteMessage,
-                      handler=self._refresh,
-                      filter=dfilter)
-
-        hub.subscribe(self, msg.DataUpdateMessage,
-                      handler=self._refresh,
-                      filter=dfilter)
-
-        hub.subscribe(self, msg.ComponentsChangedMessage,
-                      handler=self._refresh,
-                      filter=dfilter)
-
-        hub.subscribe(self, msg.NumericalDataChangedMessage,
-                      handler=self._refresh,
-                      filter=dfilter)
-
-    def unregister(self, hub):
-        super(TableViewer, self).unregister(hub)
-        hub.unsubscribe_all(self)
-
-    def _refresh(self, msg=None):
-        self._sync_layers()
-        self.model.data_changed()
-
-    def _sync_layers(self):
-
-        for layer_artist in self.layers:
-            if layer_artist.layer is not self.data and layer_artist.layer not in self.data.subsets:
-                self._layer_artist_container.remove(layer_artist)
-
-        if self.data not in self._layer_artist_container:
-            self._layer_artist_container.append(TableLayerArtist(self.data, self))
-
-        for subset in self.data.subsets:
-            if subset not in self._layer_artist_container:
-                self._layer_artist_container.append(TableLayerArtist(subset, self))
-
-    @messagebox_on_error("Failed to add data")
-    def add_data(self, data):
-        self.data = data
+    def _on_layers_changed(self, *args):
+        for layer_state in self.state.layers:
+            if isinstance(layer_state.layer, BaseData):
+                break
+        else:
+            return
+        self.data = layer_state.layer
         self.setUpdatesEnabled(False)
         self.model = DataTableModel(self)
         self.ui.table.setModel(self.model)
         self.setUpdatesEnabled(True)
-        self._sync_layers()
-        return True
 
-    def add_subset(self, subset):
-        return True
+    @messagebox_on_error("Failed to add data")
+    def add_data(self, data):
+        with self._layer_artist_container.ignore_empty():
+            self.state.layers[:] = []
+            return super(TableViewer, self).add_data(data)
+
+    @property
+    def window_title(self):
+        if len(self.state.layers) > 0:
+            return 'Table: ' + self.state.layers[0].layer.label
+        else:
+            return 'Table'
 
     def closeEvent(self, event):
         """
@@ -298,21 +287,12 @@ class TableViewer(DataViewer):
         we swap out with a tiny data set before closing
         """
         super(TableViewer, self).closeEvent(event)
-        d = Data(x=[0])
-        self.ui.table.setModel(DataTableModel(d))
+        self.model._data = Data(x=[0])
         event.accept()
 
-    def restore_layers(self, rec, context):
-        # For now this is a bit of a hack, we assume that all subsets saved
-        # for this viewer are from dataset, so we just get Data object
-        # then just sync the layers.
-        for layer in rec:
-            c = lookup_class_with_patches(layer.pop('_type'))  # noqa
-            props = dict((k, context.object(v)) for k, v in layer.items())
-            layer = props['layer']
-            if isinstance(layer, Data):
-                self.add_data(layer)
-            else:
-                self.add_data(layer.data)
-            break
-        self._sync_layers()
+    def get_layer_artist(self, cls, layer=None, layer_state=None):
+        return cls(self, self.state, layer=layer, layer_state=layer_state)
+
+    @staticmethod
+    def update_viewer_state(rec, context):
+        return update_table_viewer_state(rec, context)
